;;;; -*- Mode: Lisp -*-
;;;; $Id: compiler.lisp 540 2008-09-28 07:43:04Z binghe $

(in-package :asn.1)

(defvar *mib-name-map-table* (make-hash-table :test #'string=))

;;; Empty and old MIB module
(defvar *mib-name-map* nil)

(defun make-mib-name-map ()
  (clrhash *mib-name-map-table*)
  (mapcar #'(lambda (x)
              (setf (gethash (car x) *mib-name-map-table*) (cdr x)))
          *mib-name-map*))

(defun get-mib-name (name)
  (gethash name *mib-name-map-table* name))

(defgeneric compile-asn.1 (object &key &allow-other-keys))

(defmethod compile-asn.1 ((file t) &key to &allow-other-keys)
  (let ((result (compile-asn.1 (parse file))))
    (if (pathnamep to)
      (let ((head (merge-pathnames (make-pathname :type "lisp-expr")
                                   to))
            (*package* (find-package :asn.1))
            (*print-case* :downcase))
        (with-open-file (h head :direction :output :if-exists :supersede)
          (format h ";;;; -*- Mode: Lisp -*-~%;;;; Auto-generated from ~A~%" file)
          (pprint (remove-if #'null
                             (mapcar #'get-mib-name (car result))) h)
          (terpri h))
        (with-open-file (s to :direction :output :if-exists :supersede)
          (format s ";;;; -*- Mode: Lisp -*-~%;;;; Auto-generated from ~A~%" file)
          (dolist (i (sort-definitions (cdr result)))
            (pprint i s))
          (terpri s)))
      result)))

(defmethod compile-asn.1 ((rtl list) &key &allow-other-keys)
  (if (atom rtl)
      rtl
    (compile-asn.1-internal (car rtl) (cdr rtl))))

(defgeneric compile-asn.1-internal (type rtl))

(defmethod compile-asn.1-internal ((type symbol) (rtl list))
  (declare (ignore rtl))
  `(defunknown ,type))

(defmethod compile-asn.1-internal ((type (eql :module)) (rtl list))
  (declare (ignore type))
  (destructuring-bind (module-name xxx) rtl
    (if (null xxx)
        `(()
          (in-package :asn.1))
      (destructuring-bind ((exports imports) body) xxx
        (declare (ignore exports))
        (append `((,module-name ,@(compile-imports imports))
                  ((in-package :asn.1))
                  ((setf *current-module* ',module-name))
                  ((eval-when (:load-toplevel :execute)
                     (pushnew ',module-name *mib-modules*))))
                (mapcar #'compile-asn.1 body))))))

(defun compile-imports (rtl)
  (when rtl
    (let ((imports (cdr rtl)))
      (mapcar #'(lambda (x) (car (last x)))
              imports))))

(defmethod compile-asn.1-internal ((type (eql :value-assignment)) (rtl list))
  (declare (ignore type))
  (compile-va (car rtl) (cdr rtl)))

(defmethod compile-asn.1-internal ((type (eql :type-assignment)) (rtl list))
  (declare (ignore type))
  (compile-ta (first rtl) (second rtl)))

(defmethod compile-asn.1-internal ((type (eql :define)) (rtl list))
  (declare (ignore type))
  (compile-df (car rtl) (cdr rtl)))

(defmethod compile-asn.1-internal ((type (eql :macro)) (rtl list))
  (declare (ignore type))
  `((defmacro ,(car rtl) ())))

(defun compile-ta (name specifier)
  `((deftype ,name () ',(compile-type specifier))))

(defgeneric compile-va (type rtl))

(defmethod compile-va ((type symbol) (rtl list))
  (declare (ignore rtl))
  `((defunknown ',type)))

(defmethod compile-va ((type (eql :object-identifier)) (rtl list))
  (destructuring-bind ((name value)) rtl
    (setf value (last value 2))
    (destructuring-bind (parent value) value
      (if (consp parent)
          (setf parent (car parent)))
      `((defoid ,name (,parent ,value)
          (:type 'OBJECT-IDENTITY))
        (,name ,parent)))))

(defgeneric compile-df (type rtl))

(defmethod compile-df ((type symbol) (rtl list))
  (destructuring-bind ((name options) (parent value)) rtl
    (let ((option-arguments (compile-df-options options)))
      `((defoid ,name (,parent ,value)
          (:type ',type)
          ,@option-arguments)
        (,name ,parent)))))

(defmethod compile-df ((type (eql 'OBJECT-IDENTITY)) (rtl list))
  (destructuring-bind ((name options) (parent value)) rtl
    (let ((option-arguments (compile-df-options options)))
      `((defoid ,name (,(if (and (numberp parent)
                                 (zerop parent))
                            '|zero| parent) ,value)
          (:type 'OBJECT-IDENTITY)
          ,@option-arguments)
        (,name ,parent)))))

(defmethod compile-df ((type (eql 'TRAP-TYPE)) (rtl list))
  (declare (ignore type rtl))
  `((defunknown :trap-type)))

#|
((SYNTAX |Counter32|)
 (MAX-ACCESS |read-only|)
 (STATUS |obsolete|)
 (DESCRIPTION
  "The total number of SNMP Get-Next PDUs which have
   been generated by the SNMP protocol entity."))
-->
(:syntax integer
 :max-access :read-only
 :status :obsolete
 :description "The total number of SNMP Get-Next PDUs which have
               been generated by the SNMP protocol entity."
|#

(defun compile-df-options (options)
  (let ((description-p nil))
    (labels ((iter (o acc)
               (if (null o)
                   (nreverse acc)
                 (let ((o-item (car o)))
                   (iter (cdr o)
                         ;; make sure multi-description leave only first!
                         (let* ((compiled (compile-dfo-internal (first o-item)
                                                                (second o-item)))
                                (key (car compiled)))
                           (cond ((and (eq key :description)
                                       (not description-p))
                                  (prog1 (cons compiled acc)
                                    (setf description-p t)))
                                 ((and (eq key :description)
                                       description-p)
                                  acc)
                                 (t (cons compiled acc)))))))))
      (delete-if #'null (iter options nil)))))

(defgeneric compile-dfo-internal (key value))

(defmethod compile-dfo-internal ((key symbol) (value t))
  "Unknown df-option key"
  (declare (ignore key value))
  nil)

(defmethod compile-dfo-internal ((key (eql 'SYNTAX)) (value t))
  "Simple SYNTAX: only one symbol, map it into CL type"
  (declare (ignore key))
  `(:syntax ',(compile-type value)))

(defmethod compile-dfo-internal ((key (eql 'DESCRIPTION)) (value string))
  (declare (ignore key))
  `(:description ,value))

(defmethod compile-dfo-internal ((key (eql 'STATUS)) (value symbol))
  (declare (ignore key))
  `(:status ',value))

(defmethod compile-dfo-internal ((key (eql 'MAX-ACCESS)) (value symbol))
  (declare (ignore key))
  `(:max-access ',value))
