;;;; -*- Mode: Lisp -*-
;;;; $Id: object-id.lisp 540 2008-09-28 07:43:04Z binghe $

(in-package :asn.1)

(defvar *mib-modules* nil)
(defvar *mib-module-dependency* (make-hash-table))

(deftype access () '(member |not-accessible|
                            |accessible-for-notify|
                            |read-only|
                            |read-write|
                            |read-create|))

(deftype status () '(member |current|
                            |deprecated|
                            |mandatory|
                            |obsolete|))

(deftype oid-syntax () '(or symbol list))

(deftype oid-type ()
  '(member module-identity
           module-compliance
           object-identity
           object-type
           object-group
           notification-type
           notification-group))

(defclass object-id (asn.1-type)
  ((name        :type string
                :reader oid-name
                :initarg :name)
   (value       :type integer
                :reader oid-value
                :initarg :value)
   (type        :type oid-type
                :reader oid-type
                :initarg :type)
   (syntax      :type oid-syntax
                :reader oid-syntax
                :initarg :syntax)
   (max-access  :type access
                :reader oid-max-access
                :initarg :max-access)
   (status      :type status
                :reader oid-status
                :initarg :status)
   (description :type string
                :reader oid-description
                :initarg :description)
   (module      :type symbol
                :reader oid-module
                :initarg :module)
   (parent      :type object-id
                :reader oid-parent
                :initarg :parent)
   (children    :type hash-table
                :accessor oid-children
                :initform (make-hash-table)))
  (:documentation "OBJECT IDENTIFIER"))

(defmethod initialize-instance :after ((instance object-id)
                                       &rest initargs &key &allow-other-keys)
  (declare (ignore initargs))
  (if (and (slot-boundp instance 'parent)
           (slot-boundp instance 'name)
           (slot-boundp instance 'value))
      (setf (gethash (oid-value instance)
                     (oid-children (oid-parent instance)))
            instance)))

(defun list-children (oid)
  (declare (type object-id oid))
  (let ((all-children '()))
    (maphash #'(lambda (key value)
                 (declare (ignore key))
                 (push value all-children))
             (oid-children oid))
    all-children))

(defun delete-object (oid)
  (declare (type object-id oid))
  (remhash (oid-value oid)
           (oid-children (oid-parent oid))))

(defun oid-parent-p (oid)
  (declare (type object-id oid))
  (slot-boundp oid 'parent))

(defun oid-module-p (oid)
  (declare (type object-id oid))
  (slot-boundp oid 'module))

(defun oid-number-list (oid)
  (declare (type object-id oid))
  (labels ((iter (o acc)
             (if (slot-boundp o 'parent)
                 (iter (oid-parent o) (cons (oid-value o) acc))
               acc)))
    (iter oid nil)))

(defun oid-name-list (oid)
  (declare (type object-id oid))
  (labels ((iter (o acc)
             (if (slot-boundp o 'parent)
                 (iter (oid-parent o) (cons (if (slot-boundp o 'name)
                                                (oid-name o)
                                              (oid-value o))
                                            acc))
               acc)))
    (iter oid nil)))

(defun oid-length (oid)
  (declare (type object-id oid))
  (list-length (oid-number-list oid)))

(defmethod plain-value ((object object-id))
  (coerce (oid-number-list object) 'vector))

(defmethod ber-equal ((a object-id) (b object-id))
  (or (equal (oid-number-list a) (oid-number-list b))
      (and (equal (oid-number-list a) '(0)) (equal (oid-number-list b) '(0 0)))
      (and (equal (oid-number-list b) '(0)) (equal (oid-number-list a) '(0 0)))))

;;; BER Encode & Decode (:object-identifier)
(defmethod ber-encode ((value object-id))
  (labels ((number-get (n)
             (if (zerop n) (values (list 0) 1)
               (number-split n 0 nil 0)))
           (number-split (n p acc l)
             (if (zerop n) (values acc l)
               (multiple-value-bind (q r) (floor n 128)
                 (number-split q 1 (cons (logior (ash p 7) r) acc) (1+ l)))))
           (iter (oids acc len)
             (if (endp oids)
                 (values acc len)
               (multiple-value-bind (sub-oid sub-length) (number-get (car oids))
                 (iter (cdr oids) (nconc acc sub-oid) (+ len sub-length))))))
    (let ((subids (oid-number-list value)))
      (multiple-value-bind (v l)
          (case (oid-length value)
            (0 (values nil 0))
            (1 (number-get (* (first subids) 40)))
            (2 (number-get (+ (* (first subids) 40) (second subids))))
            (otherwise (apply #'iter
                              (cddr subids)
                              (multiple-value-list
                               (number-get (+ (* (first subids) 40)
                                              (second subids)))))))
        (concatenate 'vector
                     (ber-encode-type 0 0 6)
                     (ber-encode-length l)
                     v)))))

;;; FIXME: #(6 1 40), |iso| should be decoded to 1, not 1.0
(defmethod ber-decode-value ((s stream) (type (eql :object-identifier)) length)
  (declare (type stream s)
           (type fixnum length)
           (ignore type))
  (if (zerop length)
      (oid '(0))
    (labels ((get-number (acc len)
               (let* ((byte (read-byte s))
                      (val (logior (ash acc 7) (logand byte 127))))
                 (if (< byte 128) (values val len)
                   (get-number val (1+ len)))))
             (iter (left-length acc head-p)
               (declare (type fixnum left-length)
                        (type list acc))
               (if (zerop left-length)
                   (nreverse acc)
                 (multiple-value-bind (n l) (get-number 0 1)
                   (if head-p
                       (multiple-value-bind (q r) (floor n 40)
                         (iter (- left-length l) (cons r (cons q acc)) nil))
                     (iter (- left-length l) (cons n acc) nil))))))
      (oid (iter length nil t)))))

(eval-when (:load-toplevel :execute)
  (install-asn.1-type :object-identifier 0 0 6))

;;; Object ID utilities, from Lisp-SNMP
(defun oid-< (oid-1 oid-2)
  "test if oid-1 is oid-2's child"
  (let ((o-1 (nreverse (oid-number-list oid-1)))
        (o-2 (nreverse (oid-number-list oid-2)))
        (o-1-len (oid-length oid-1))
        (o-2-len (oid-length oid-2)))
    (if (<= o-1-len o-2-len) nil
      (equal o-2 (nthcdr (- o-1-len o-2-len) o-1)))))

(defun oid->= (oid-1 oid-2)
  (not (oid-< oid-1 oid-2)))

(defvar *root-object-id*
  (make-instance 'object-id))

(defvar |zero|
  (make-instance 'object-id
                 :name "zero"
                 :value 0
                 :parent *root-object-id*))

(defvar |iso|
  (make-instance 'object-id
                 :name "iso"
                 :value 1
                 :parent *root-object-id*))

(defmacro defoid (name (parent value) &body options)
  `(defvar ,name
     ,(append `(make-instance 'object-id
                              :name (symbol-name ',name)
                              :value ,value
                              :parent ,parent
                              :module *current-module*)
              (apply #'append options))))
